import numpy
import scipy

import msfit
import msmath
import mssputil

from matplotlib.figure import Figure

CALIB_STR=('NONE','RLS','RANSAC','LOWESS')
CALIB_NONE=-1
CALIB_RLS=0
CALIB_RANSAC=1
CALIB_LOWESS=2
CALIB_LS=3

ESTIM_STR=('NONE','SD','ROBUST SD')
ESTIM_NONE=-1
ESTIM_SD=0

ESTIM_MAD=1

# --------------------------------------------------------------------------
#
def get_calibration_type(name):
	calibration_type=-1
	for i in range(0,len(CALIB_STR)):
		if name.lower()==CALIB_STR[i].lower():
			calibration_type=i-1
			break
	return calibration_type

# --------------------------------------------------------------------------
#
def get_estimation_type(name):
	estimation_type=-1
	for i in range(0,len(ESTIM_STR)):
		if name.lower()==ESTIM_STR[i].lower():
			estimation_type=i-1
			break
	return estimation_type

# --------------------------------------------------------------------------
#
def get_calibrated_rt(rt,spectrum):
	
	cal_rt=rt+mssputil.get_rt_correction(spectrum,rt)
	
	rt_range=mssputil.get_rt_range(spectrum,False)
	cal_rt=min(max(cal_rt,rt_range[0]),rt_range[1])

	return cal_rt

# --------------------------------------------------------------------------
#
def get_calibrated_mz():
	
	return mz

# --------------------------------------------------------------------------
#
def get_mz_delta(peptide_list,mz_correction=False,spectrum=None):

	delta_mz=[]
	mz=[]
	for i in range(len(peptide_list)):
		peptide=peptide_list[i]
		for j in range(len(peptide.query_list)):
			query=peptide.get(j)
			if query.present==1:
				if mz_correction==1:
					dmz_correction=mssputil.get_mz_correction_ppm(spectrum,query.cmz)
				else:
					dmz_correction=0
				dmz=query.mz-query.cmz
				if query.cmz!=0 and dmz<1:
					delta_mz.append(1e+6*dmz/query.cmz-dmz_correction)
					mz.append(query.cmz)
	
	return mz,delta_mz

# --------------------------------------------------------------------------
#
def get_rt_delta(peptide_list,rt_correction=False,spectrum=None):
	delta_rt=[]
	rt=[]
	for i in range(len(peptide_list)):
		peptide=peptide_list[i]
		for j in range(len(peptide.query_list)):
			query=peptide.get(j)
			if query.present==1 and query.massref_rt!=0:
				if rt_correction==True and spectrum!=None:
					drt_correction=mssputil.get_rt_correction(spectrum,query.massref_rt)
				else:
					drt_correction=0
				delta_rt.append(query.rt-(query.massref_rt+drt_correction))
				rt.append(query.massref_rt)
						
	return rt,delta_rt

# --------------------------------------------------------------------------
#
def get_rp_delta(peptide_list):
	rp=[]
	mz=[]
	for i in range(len(peptide_list)):
		peptide=peptide_list[i]
		for j in range(len(peptide.query_list)):
			query=peptide.get(j)
			peak=query.get(0)
			fits=msfit.get_fits(peak)
			if query.present==1 and fits!=None:
				fit=fits[peak.best_fit]
				(fit_probs,fit_mzs,fit_sigmas)=msfit.mz_fit_peaks_form_params(fit.mz_params)
				index=numpy.argmax(fit_probs)
				if fit_mzs[index]>0 and fit_sigmas[index]>0:
					rp.append(fit_mzs[index]/(fit_sigmas[index]*2*numpy.sqrt(2*numpy.log(2))))
					mz.append(query.cmz)
	return mz,rp

# --------------------------------------------------------------------------
#
def calibrate(session,points,points_delta,calib_method,estim_method,estim_m):

	a=0
	b=0
	sf=None
	sa=None
	ransac_succes=False
	std=0

	if calib_method==CALIB_LS:
		a,b,inliers_index=msmath.rls_linfit(points,points_delta,0)

		corrected_delta=points_delta-(a*points+b)

	if calib_method==CALIB_RLS:
		a,b,inliers_index=msmath.rls_linfit(points,points_delta,1)

		corrected_delta=points_delta-(a*points+b)


	elif calib_method==CALIB_RANSAC:

		min_n=len(points_delta)/4
		min_n_inliers=len(points_delta)/2
		iterations=1000
		error_th=30

		a,b,inliers_index,ransac_succes=msmath.ransac_linfit(points,points_delta,min_n,min_n_inliers,error_th,iterations)
		if ransac_succes==False:
			a,b,inliers_index=msmath.rls_linfit(points,points_delta,1)

		corrected_delta=points_delta-(a*points+b)

	elif calib_method==CALIB_LOWESS:
		idx=numpy.argsort(points)
		points=points[idx]
		points_delta=points_delta[idx]

		fit,inliers_index=msmath.lowess(session,points,points_delta,0.4,2,0.0,6.0)

		#tu musi byc ograniczenie w przypadku gdy malo punktow!!!!!!
		sa=scipy.stats.mstats.mquantiles(points,prob=numpy.arange(0,1.05,0.05))

		sf=numpy.zeros(len(sa))
		for i in range(0,len(sa)):
			i1=(points<=sa[i]).nonzero()
			sf[i]=fit[i1[0][-1]]

		corrected_delta=numpy.zeros(len(points))
		for i in range(0,len(points)):
			corrected_delta[i]=points_delta[i]-msmath.lowess_val(sa,sf,points[i])

	if estim_method==ESTIM_MAD:
		std=estim_m*msmath.std_mad(corrected_delta[inliers_index])
	elif estim_method==ESTIM_SD:
		std=estim_m*numpy.std(corrected_delta[inliers_index])

	return points,points_delta,a,b,sf,sa,std,inliers_index,ransac_succes


# --------------------------------------------------------------------------
#
def calibration_plot(xlabel,ylabel,points,points_delta,a,b,sf,sa,std,inliers_index,ransac_succes=False,ranges=None):
	
	fig=Figure()
	p=fig.add_subplot(111)

	p.plot(points,points_delta,'ro')
	p.plot(points[inliers_index],points_delta[inliers_index],'go',label='inliers')

	if a!=0 or b!=0:
		if ransac_succes==True:
			color='b'
		else:
			color='k'

		x=numpy.sort(points)
		p.plot(x,a*x+b,color=color,linewidth=2)
		p.plot(x,a*x+b+std,linewidth=1,color=color,linestyle='--')
		p.plot(x,a*x+b-std,linewidth=1,color=color,linestyle='--')

		ax=p.axis()
		p.text(ax[1],ax[2],('a=%0.4f b=%0.4f ' % (a,b)),horizontalalignment='right',verticalalignment='bottom',color=color)


	elif sa!=None and sf!=None:
		p.plot(sa,sf,'ko')
		p.plot(sa,sf,color='k',linewidth=2)
		p.plot(sa,sf+std,color='k',linewidth=2,linestyle='--')
		p.plot(sa,sf-std,color='k',linewidth=2,linestyle='--')

	p.set_xlabel(xlabel)
	p.set_ylabel(ylabel)
	if ranges!=None:
		p.set_ylim((-ranges[0],ranges[1]))
		
	return fig

# --------------------------------------------------------------------------
#
def histogram_plot(xlabel,ylabel,points_delta):
	
	fig=Figure()
	p=fig.add_subplot(111)
	
	p.hist(points_delta,100)
	p.set_xlabel(xlabel)
	p.set_ylabel(ylabel)
	
	return fig

# --------------------------------------------------------------------------
#
def rp_calibration_params_plot(spectrum):
	
	fig=Figure()
	p=fig.add_subplot(111)

	mz=numpy.array(mssputil.get_mz_range(spectrum))
	a=spectrum.rp_a
	b=spectrum.rp_b
	
	p.plot(mz,a*mz+b,color='k',linewidth=2)

	ax=p.axis()
	p.text(ax[1],ax[2],('a=%0.4f b=%0.4f ' % (a,b)),horizontalalignment='right',verticalalignment='bottom',color='k')

	p.set_xlabel('m/z')
	p.set_ylabel('RP')
	
	return fig


# --------------------------------------------------------------------------
#
def mz_calibration_params_plot(spectrum):
	
	fig=Figure()
	p=fig.add_subplot(111)

	mz=numpy.array(mssputil.get_mz_range(spectrum))
	a=spectrum.mz_correction_a
	b=spectrum.mz_correction_b
	std=spectrum.mz_precision
	
	p.plot(mz,a*mz+b,color='k',linewidth=2)
	p.plot(mz,a*mz+b+std,linewidth=1,color='k',linestyle='--')
	p.plot(mz,a*mz+b-std,linewidth=1,color='k',linestyle='--')

	ax=p.axis()
	p.text(ax[1],ax[2],('a=%0.4f b=%0.4f ' % (a,b)),horizontalalignment='right',verticalalignment='bottom',color='k')

	p.set_xlabel('m/z')
	p.set_ylabel('m/z error [ppm]')
	
	return fig

# --------------------------------------------------------------------------
#
def rt_calibration_params_plot(spectrum):
	
	fig=Figure()
	p=fig.add_subplot(111)

	std=spectrum.mz_precision/2

	if spectrum.rt_correction_method==CALIB_LOWESS and spectrum.rt_correction_sf!=None and spectrum.rt_correction_sa!=None:
		sf=numpy.array(spectrum.rt_correction_sf)
		sa=numpy.array(spectrum.rt_correction_sa)
					
		p.plot(sa,sf,'ko')
		p.plot(sa,sf,color='k',linewidth=2)
		p.plot(sa,sf+std,color='k',linewidth=2,linestyle='--')
		p.plot(sa,sf-std,color='k',linewidth=2,linestyle='--')
	else:
		rt=numpy.array(mssputil.get_rt_range(spectrum))
		a=spectrum.rt_correction_a
		b=spectrum.rt_correction_b
	
		p.plot(rt,a*rt+b,color='k',linewidth=2)
		p.plot(rt,a*rt+b+std,linewidth=1,color='k',linestyle='--')
		p.plot(rt,a*rt+b-std,linewidth=1,color='k',linestyle='--')
		
		ax=p.axis()
		p.text(ax[1],ax[2],('a=%0.4f b=%0.4f ' % (a,b)),horizontalalignment='right',verticalalignment='bottom',color='k')

	p.set_xlabel('RT')
	p.set_ylabel('RT error [min]')
	
	return fig
