import time
import math
import numpy
import scipy
import scipy.stats
import scipy.optimize
import scipy.special
import scipy.integrate

import matplotlib
from matplotlib.figure import Figure

import ms
import msmath
import msfilter
import mspeak
import mstkutil
import mssputil


MS_FIT_PARAMS=21
MS_RT_FIT_PARAMS=4
MS_MZ_FIT_PARAMS=3

RT_FIT_EMG=-100.0
RT_FIT_GAUSS=-200.0
RT_FIT_PMG=-300.0

VALUE_DATA_HEIGHT=0
VALUE_FIT_HEIGHT=1
VALUE_FIT_VOLUME=2
VALUE_SFIT_HEIGHT=3
VALUE_SFIT_VOLUME=4


# ----------------------------------------------------------------------------
#
class value_menu(mstkutil.option_menu):

	def __init__(self,parent,label,allow_no_choice=0,with_rt=0):
		self.allow_no_choice=allow_no_choice
		self.with_rt=with_rt

		if self.with_rt!=0:
			mstkutil.option_menu.__init__(self,parent,label,('Retention time','Data height','Fit height','Fit volume','Sparky fit height','Sparky fit volume'),'Fit height')
		else:
			mstkutil.option_menu.__init__(self,parent,label,('Data height','Fit height','Fit volume','Sparky fit height','Sparky fit volume'),'Fit height')

	# --------------------------------------------------------------------------
	#
	def get_value(self):
		entry=self.get()

		if entry=='Data height':
			value=VALUE_DATA_HEIGHT
		elif entry=='Fit height':
			value=VALUE_FIT_HEIGHT
		elif entry=='Fit volume':
			value=VALUE_FIT_VOLUME
		elif entry=='Sparky fit height':
			value=VALUE_SFIT_HEIGHT
		elif entry=='Sparky fit volume':
			value=VALUE_SFIT_VOLUME
		else:
			value=-1

		return value
	# --------------------------------------------------------------------------
	#
	def set_value(self,value):
		entry='Fit height'

		if value==VALUE_DATA_HEIGHT:
			entry='Data height'
		elif value==VALUE_FIT_HEIGHT:
			entry='Fit height'
		elif value==VALUE_FIT_VOLUME:
			entry='Fit volume'
		elif value==VALUE_SFIT_HEIGHT:
			entry='Sparky fit height'
		elif value==VALUE_SFIT_VOLUME:
			entry='Sparky fit volume'
		elif value<0 and self.with_rt!=0:
			entry='Retention time'

		entry=self.set(entry)

# ----------------------------------------------------------------------------
#
class MSFit:
	
	def __init__(self):
							#fitsCount
		self.peaks_count=0	#peaksCount

		self.fit_nr=0		#fitNr

		self.rt_pos=0		#rtPos
		self.rt_params=[]	#rtParam1,rtParam2,rtParam3,rtParam4
		self.rt_flag=0		#rtFlag
		self.rt_rmse=0		#rtMSE
		self.rt_rsquare=0	#rtRSquare
		self.rt_chi=0		#rtChi
		self.rt_area=0		#rtArea

		self.mz_pos=0		#mzPos
		self.mz_delta=0		#mzDelta
		self.mz_params=[]	#mzParam1.1,mzParam1.2,mzParam1.3,mzParam2.1,mzParam2.2,mzParam2.3,...
		self.mz_flag=0		#mzFlag
		self.mz_rmse=0		#mzMSE
		self.mz_rsquare=0	#mzRSquare
		self.mz_chi=0		#mzChi		#tymczasowo uzywane do przechowywania liczby dopasowanych pikow
		self.mz_area=0		#mzArea
		
		self.mz_env_rmse=0	#mzEnvMSE
		self.mz_env_rsquare=0	#mzEnvRSquare
		self.mz_env_chi=0	#mzEnvChi

		self.fit_rmse=0		#fitMSE
		self.fit_rsquare=0	#fitRSquare
		self.fit_chi=0		#fitChi		#tymczasowo uzywane do przechowywania wysokosci piku chromatogramu???
		self.fit_volume=0	#fitVolume

	# ----------------------------------------------------------------------------
	#
	def set_mz(self,mz,spectrum):
		if spectrum.domain==mssputil.DOMAIN_FREQ:
			self.mz_pos=mssputil.mz2freq_s(mz,spectrum)
		else:
			self.mz_pos=-1*mz

	# ----------------------------------------------------------------------------
	#
	def get_mz(self,spectrum):
		if spectrum.domain==mssputil.DOMAIN_FREQ:
			mz=mssputil.freq2mz_s(self.mz_pos,spectrum)
		else:
			mz=-1*self.mz_pos

		return(mz)
	
	# ----------------------------------------------------------------------------
	#
	def get_fit_height(self,peak_index=-1):
		if peak_index!=-1:
			h=self.get_peak_height(peak_index)
		else:
			h=0
			for i in range(0,self.peaks_count):
				h=h+self.get_peak_height(i)
		
		return h
	
	# ----------------------------------------------------------------------------
	#
	def get_peak_height(self,peak_index=-1):
		h=0
		if peak_index>=0 and peak_index<self.peaks_count:
			index=MS_MZ_FIT_PARAMS*peak_index
			h=self.mz_params[index]
		
		return h
	
	# ----------------------------------------------------------------------------
	#
	def get_fit_volume(self,peak_index=-1):
		if peak_index!=-1:
			v=self.get_peak_volume(peak_index)
		else:
			v=0
			for i in range(0,self.peaks_count):
				v=v+self.get_peak_volume(i)
		return(v)
		
	# ----------------------------------------------------------------------------
	#
	def get_peak_volume(self,peak_index=-1):
		#to dziala tylko dla gaussowskich		
		v=0
		if peak_index>=0 and peak_index<self.peaks_count:
			index=MS_MZ_FIT_PARAMS*peak_index
			A=self.mz_params[index]
			sigma_x=self.mz_params[index+2]
			sigma_y=self.rt_params[2]
		
			v=2*numpy.pi*A*sigma_x*sigma_y
		
			#print '%f, %f, %f, %f' % (A,sigma_x,sigma_y,v)
		return v

# ----------------------------------------------------------------------------
#
def set_fits(peak,peak_fits):

	fits_count=len(peak_fits)

	if fits_count>0:
		peaks_count=peak_fits[0].peaks_count
		params=numpy.zeros((2+fits_count*(MS_FIT_PARAMS+MS_RT_FIT_PARAMS+peaks_count*MS_MZ_FIT_PARAMS)))

		params[0]=fits_count
		params[1]=peaks_count

		counter=2
		for i in range(0,fits_count):
			peak_fit=peak_fits[i]

			params[counter]=peak_fit.fit_nr
			counter=counter+1
			params[counter]=peak_fit.rt_pos
			counter=counter+1
			for j in range(0,MS_RT_FIT_PARAMS):
				params[counter]=peak_fit.rt_params[j]
				counter=counter+1
			params[counter]=peak_fit.rt_flag
			counter=counter+1
			params[counter]=peak_fit.rt_rmse
			counter=counter+1
			params[counter]=peak_fit.rt_rsquare
			counter=counter+1
			params[counter]=peak_fit.rt_chi
			counter=counter+1
			params[counter]=peak_fit.rt_area
			counter=counter+1

			params[counter]=peak_fit.mz_pos
			counter=counter+1
			params[counter]=peak_fit.mz_delta
			counter=counter+1
			for i in range(0,peaks_count):
				for j in range(0,MS_MZ_FIT_PARAMS):
					params[counter]=peak_fit.mz_params[i*MS_MZ_FIT_PARAMS+j]
					counter=counter+1
					
			params[counter]=peak_fit.mz_flag
			counter=counter+1
			params[counter]=peak_fit.mz_rmse
			counter=counter+1
			params[counter]=peak_fit.mz_rsquare
			counter=counter+1
			params[counter]=peak_fit.mz_chi
			counter=counter+1
			params[counter]=peak_fit.mz_area
			counter=counter+1

			params[counter]=peak_fit.mz_env_rmse
			counter=counter+1
			params[counter]=peak_fit.mz_env_rsquare
			counter=counter+1
			params[counter]=peak_fit.mz_env_chi
			counter=counter+1

			params[counter]=peak_fit.fit_rmse
			counter=counter+1
			params[counter]=peak_fit.fit_rsquare
			counter=counter+1
			params[counter]=peak_fit.fit_chi
			counter=counter+1
			params[counter]=peak_fit.fit_volume
			counter=counter+1
			
		params=params.tolist()
	else:
		params=[]
	peak.fits_params=params

# ----------------------------------------------------------------------------
#
def get_fits(peak):
	peak_fits=[]

	params=peak.fits_params
	if params!=None:
		fits_count=int(params[0])
		peaks_count=int(params[1])

		counter=2
		for i in range(0,fits_count):
			peak_fit=MSFit()
			peak_fit.peaks_count=peaks_count

			peak_fit.fit_nr=params[counter]
			counter=counter+1

			peak_fit.rt_pos=params[counter]
			counter=counter+1
			peak_fit.rt_params=numpy.zeros(MS_RT_FIT_PARAMS)
			for k in range(0,MS_RT_FIT_PARAMS):
				peak_fit.rt_params[k]=params[counter]
				counter=counter+1
			peak_fit.rt_flag=params[counter]
			counter=counter+1
			peak_fit.rt_rmse=params[counter]
			counter=counter+1
			peak_fit.rt_rsquare=params[counter]
			counter=counter+1
			peak_fit.rt_chi=params[counter]
			counter=counter+1
			peak_fit.rt_area=params[counter]
			counter=counter+1

			peak_fit.mz_pos=params[counter]
			counter=counter+1
			peak_fit.mz_delta=params[counter]
			counter=counter+1
			peak_fit.mz_params=numpy.zeros(peaks_count*MS_MZ_FIT_PARAMS)
			for j in range(0,peaks_count):
				for k in range(0,MS_MZ_FIT_PARAMS):
					peak_fit.mz_params[j*MS_MZ_FIT_PARAMS+k]=params[counter]
					counter=counter+1
			peak_fit.mz_flag=params[counter]
			counter=counter+1
			peak_fit.mz_rmse=params[counter]
			counter=counter+1
			peak_fit.mz_rsquare=params[counter]
			counter=counter+1
			peak_fit.mz_chi=params[counter]
			counter=counter+1
			peak_fit.mz_area=params[counter]
			counter=counter+1

			peak_fit.mz_env_rmse=params[counter]
			counter=counter+1
			peak_fit.mz_env_rsquare=params[counter]
			counter=counter+1
			peak_fit.mz_env_chi=params[counter]
			counter=counter+1

			peak_fit.fit_rmse=params[counter]
			counter=counter+1
			peak_fit.fit_rsquare=params[counter]
			counter=counter+1
			peak_fit.fit_chi=params[counter]
			counter=counter+1
			peak_fit.fit_volume=params[counter]
			counter=counter+1

			peak_fits.append(peak_fit)

	return(peak_fits)

# --------------------------------------------------------------------------
#
def get_fit_value(peak,value_type=VALUE_FIT_HEIGHT,fit=None,monoisotopic_only=1):

	value=0

	#usunac inne niz fitu
	#przerobic wartosci w mssputil.get_peak_height
	if value_type==VALUE_DATA_HEIGHT:
		value=mssputil.get_peak_height(peak,1)
	elif value_type==VALUE_SFIT_HEIGHT:
		value=mssputil.get_peak_height(peak,0)
	elif value_type==VALUE_SFIT_VOLUME:
		value=mssputil.get_peak_height(peak,2)
	else:
		if fit==None:
			fits=get_fits(peak)
			if len(fits)>0:
				fit=fits[peak.best_fit]

		if fit!=None:
			if value_type==VALUE_FIT_HEIGHT:
				if monoisotopic_only==1:
					value=fit.get_fit_height(0)	#powinno byc od najwyzszego
				else:
					value=fit.get_fit_height(-1)
			elif value_type==VALUE_FIT_VOLUME:
				if monoisotopic_only==1:
					value=fit.get_fit_volume(0) #powinno byc od najwyzszego
				else:
					value=fit.get_fit_volume(-1)
	return(value)

#
# Fitowanie pikow chromatogramu
#
# --------------------------------------------------------------------------
#
def fit_chromatogram_peaks(spectrum,chromat,isodistrib,charge,peaks_count,rp,mz_correction=0,mz_margin=[20,20],fit_monoisotopic_rt=False,env_rsquare_threshold=0.1,sequence=None,plot=False,do_full_fit=0,scans_margin=0):

	fits=[]
	msg=''
	
	for i in range(0,len(chromat.peaks)):
		peak=chromat.peaks[i]
		rt_range=(peak[5],peak[6],peak[4])	#(right_min,left_min,peak_max)

		(fit,mz_points,mz_data,mz_fit_data,mz_envelope_data,mz_e,mz_max_range,mz_max_intensity,mz_peaks,mz_noise,rt_points,rt_data,rt_fit_data,rt_fit_range,rt_max,rt_peaks,rt_noise,fit_success,fit_msg)=create_fit(spectrum,i+1,rt_range,isodistrib,float(charge),peaks_count,rp,mz_correction,mz_margin,fit_monoisotopic_rt,do_full_fit,scans_margin)
		msg=msg+fit_msg
		if fit_success>0:
			if math.fabs(fit.mz_env_rsquare)>env_rsquare_threshold:	#na razie abs (jak deizotopizacja bedzie OK, to mozna abs usunac)
				fit.fit_chi=peak[3]									#po co to
				fits.append(fit)
			else:
				msg=msg+(('\n ERROR: m/z R-Square below acceptance threshold (%f) ') % env_rsquare_threshold)

		if plot==True:
			if mz_envelope_data==None:
				#(mz_envelope_data,mz_envelope_points)=isodistrib.get_envelope_for_points(mz_points,charge,rp,mz_correction,mz_max_intensity)
				(mz_envelope_data,mz_envelope_points)=isodistrib.get_envelope_for_points(mz_points,charge,rp,mz_correction,mz_data.max())
			plot_chromatogram_fit(spectrum.session,chromat,i,mz_points,mz_data,mz_fit_data,mz_envelope_data,rt_points,rt_data,rt_fit_data,rt_fit_range,('%s (+%d) [fit: %d]' % (sequence,charge,i+1)))

	return fits,msg

# --------------------------------------------------------------------------
#
def create_fit(spectrum,fit_nr,rt_range,isodistrib,charge,peaks_count,rp,mz_correction,mz_margin,fit_monoisotopic_rt,do_full_fit,scans_margin,force_fit=False):

	#fit
	fit=None
	fit_msg=''
	fit_success=0

	#dane m/z
	mz_points=None
	mz_data=None
	fit_mz_data=None
	envelope_mz_data=None
	mz_e=0
	mz_max=0
	mz_max_intensity=0
	mz_max_range=None
	mz_peaks=None
	mz_noise=0

	#dane RT
	rt_points=None
	rt_data=None
	rt_fit_data=None
	rt_fit_range=None
	rt_max=0
	rt_peaks=None
	rt_noise=0

	#pobranie najwyzszych pikow obwiedni: isopeaks=[index; m/z; prob] uporzadkowane od najwyzszego do najnizszego
	isopeaks=isodistrib.get_highest_peaks(peaks_count,charge)
	mono_isopeak_mz=isodistrib.get_peak_mz(0,charge)
	highest_isopeak=isopeaks[0,0]
	highest_isopeak_mz=isopeaks[1,0]
	
	#tworzenie fitu
	fit=MSFit()
	fit.fit_nr=fit_nr
	fit.peaks_count=peaks_count	#powinno byc ustawiane w create_mz_fit

	# pobranie danych RT i znalezienie maksymalnego punktu => zrobic dwie osobne funkcje do brania trace i liczenia max w zlaznosci od wyboru czy max czy mean
	rt_fit_mz=highest_isopeak_mz
	(rt_points,rt_data,rt_max,rt_max_index,rt_traces,rt_traces_mz_points)=get_rt_trace(spectrum,rt_range,rt_fit_mz,mz_correction,mz_margin)

	fit_msg=fit_msg+('\n\n Fit %d\n Mono peak m/z: %f\n Highest peak (%d) m/z: %f\n RT fit m/z: %f\n RT fit max: %f' % (fit.fit_nr,mono_isopeak_mz,highest_isopeak,highest_isopeak_mz,rt_fit_mz,-1*rt_max))

	#fit m/z
	(mz_points,mz_data,fit_mz_data,envelope_mz_data,mz_e,mz_max,mz_max_range,mz_max_intensity,mz_peaks,mz_noise,mz_success,mz_msg)=create_mz_fit(spectrum,fit,-1*rt_max,isodistrib,charge,isopeaks,rp,mz_correction,mz_margin,scans_margin,force_fit)
	fit_msg=fit_msg+mz_msg

	if mz_success>0:
		fit_success=1
		if do_full_fit==1:
			#fit RT
			#czy nie powinno byc jeszcze raz wziete RT trace - tym razem dla faktycznego maximum i zakresu m/z wynikajacego z szerokosci piku????
			(rt_points,rt_data,rt_fit_data,rt_fit_range,rt_max,rt_peaks,rt_noise,rt_success,rt_msg)=create_rt_fit(spectrum,fit,rt_points,rt_traces,rt_traces_mz_points,rt_max,rt_max_index)
			fit_msg=fit_msg+rt_msg			
			
			#if rt_seccess>0
			#	fit_secces=1
		else:
			fit.rt_pos=rt_max
			fit.rt_params=[0,0,0,0]

	return(fit,mz_points,mz_data,fit_mz_data,envelope_mz_data,mz_e,mz_max_range,mz_max_intensity,mz_peaks,mz_noise,rt_points,rt_data,rt_fit_data,rt_fit_range,rt_max,rt_peaks,rt_noise,fit_success,fit_msg)



# --------------------------------------------------------------------------
#
def sort_fits(peak_fits,sort_type):
	fits_count=len(peak_fits)
	if fits_count>0:
		if fits_count>1:
			peak_fits.sort(lambda fit1,fit2: cmp(-1*fit1.mz_env_rsquare,-1*fit2.mz_env_rsquare))
	
		for i in range(0,len(peak_fits)):
			peak_fits[i].fit_nr=i+1
	
#
# Fitowanie po m/z
#

# --------------------------------------------------------------------------
# Tworzenie i ocena fitu m/z
def create_mz_fit(spectrum,fit,rt_max,isodistrib,charge,isopeaks,rp,mz_correction,mz_margin,scans_margin,force_fit=False):
	msg=''
	success=1
	mz_points=[]		#sprawdzic czy zmiana z None nie powoduje bledow (np. przy rysowaniu)
	mz_data=[]
	fit_mz_data=[]
	envelope_mz_data=[]
	mz_max=0
	mz_max_range=[]
	mz_max_intensity=1
	
	peaks_count=isopeaks.shape[1]
	highest_isopeak=isopeaks[0,0]
	highest_isopeak_mz=isopeaks[1,0]
	min_mz_isopeak=numpy.min(isopeaks[0,:])
	
	#niekosekwencja: raz uzywany mz_margin a raz margines wynikajacy z RP
	rp_mz_margin=highest_isopeak_mz/rp
	
	# pobieranie danych z pelnego zakresu m/z i wstepna ocena sensownosci pozycji
	(mz_e,full_mz_data,full_mz_points,mz_peaks,mz_noise,mz_e_succes,mz_e_msg)=evaluate_position(spectrum,rt_max,isodistrib,charge,[2,1/charge],mz_correction,(2*rp_mz_margin,2*rp_mz_margin),scans_margin)
	msg=msg+mz_e_msg
				
	if mz_e_succes==0:
		msg=msg+'\n ERROR: unable to compute m/z fit'
		success=0
	else:
		# pobieranie danych z zakresu m/z w ktorym bedzie liczone dopasowanie
		mz_range=isodistrib.get_peaks_mz_range(charge,(min_mz_isopeak,min_mz_isopeak+peaks_count-1),(0.5,0.5),mz_correction)
		(mz_data,mz_points,mz_index)=get_data_range(full_mz_data,full_mz_points,mz_range)
	
		#interpolacja i filtracja danych m/z
		(mz_data,mz_points,mz_step)=msmath.interpolate(mz_data,mz_points,3)
		mz_data=msfilter.filter_data(mz_data,filter_type=msfilter.FILTER_GAUSS,filter_span=1,filter_order=0)
	
		msg=msg+('\n m/z fit range: %f - %f' % (mz_range[0],mz_range[1]))
		msg=msg+('\n Max intesity in full envelope m/z range (%f - %f): %f' % (mz_range[0],mz_range[1],mz_data.max()))
	
		#wyznaczenie maksimum w zakresie najwyzszego piku !!!!!!!!!tutaj musi byc tak samo jak w chromatogramie=> poprawic w chromatogramie!!!!!!!!
		mz_max_range=isodistrib.get_peak_mz_abs_range(charge,highest_isopeak,(2*rp_mz_margin,2*rp_mz_margin),mz_correction)
		(mz_max_intensity,mz_max,mz_max_index)=get_data_max(mz_data,mz_points,mz_max_range)
	
		if numpy.isfinite(mz_max_intensity) and mz_max_intensity!=0:
			msg=msg+('\n Max intesity in highest peak m/z range (%f - %f): %f' % (mz_max_range[0],mz_max_range[1],mz_max_intensity))
	
			#!!!!!!!!!wiemy jakie jest przesuniecie => zamiast mz_correction roznica w m/z dla najwyzszych!!!!!!!!!!!!!!
			#mz_delta=polozenieteoretyczne-mz_max
			mz_fit_p,mz_fit_order,fit_probs,fit_mzs,fit_sigmas,fit_max_index,fit_peaks_count,mz_succes,mz_msg=mz_fit(mz_data,mz_points,isopeaks,rp,mz_correction,mz_margin,mz_max_intensity)
			msg=msg+mz_msg
			
			if mz_succes<0 and force_fit==True:
				mz_fit_p,mz_fit_order,fit_probs,fit_mzs,fit_sigmas,fit_max_index,fit_peaks_count,mz_succes,mz_msg=force_mz_fit(mz_data,mz_points,isopeaks,rp,mz_correction,rp_mz_margin,mz_max_intensity)
				msg=msg+mz_msg


			if mz_succes<0:
				msg=msg+'\n ERROR: unable to compute m/z fit'
				success=0;
			else:
				fit.mz_delta=numpy.around(fit_mzs[fit_max_index]-highest_isopeak_mz,6)
	
				mono_isopeak_mz=isodistrib.get_peak_mz(0,charge)
				fit.set_mz(numpy.around(mono_isopeak_mz+fit.mz_delta,6),spectrum)
	
				fit.mz_params=numpy.around(mz_fit_p,6)
				fit.mz_flag=mz_succes
				fit.mz_chi=fit_peaks_count
				
				#!!!!!!!!!w evaluate dopasowywac sie do rzeczywistej szerokosci pikow z fitu??????????
				fit_mz_data,envelope_mz_data=evaluate_mz_fit(fit,isodistrib,charge,rp,mz_data,mz_points,fit_probs,fit_mzs,fit_sigmas)
				fit.mz_env_rsquare=mz_e*fit.mz_env_rsquare
	
				#przy obecnych zabezpieczeniach (zamianie na dodatnie) to ostrzezenie nie powinno sie pojawiac
				if fit.mz_area<0:	
					msg=msg+'\n WARNING: m/z area is negative'
					msg=msg+('\n m/z fit params: %s' % str(mz_fit_p))
	
				msg=msg+('\n m/z fit max: %f\n m/z fit delta: %f\n m/z fit R-square: %f' % (fit_mzs[fit_max_index],mssputil.da2ppm(fit.mz_delta-mz_correction,highest_isopeak_mz),fit.mz_env_rsquare))
		else:
			success=0;
			if mz_max_intensity==0:
				mz_max_intensity=mz_data.max()
				msg=msg+'\n ERROR: zero intesity of the highest peak'
			else:
				mz_max_intensity=full_mz_data.max()
				msg=msg+'\n ERROR: no points in highest peak range'
	
			if mz_max_intensity==0:
				mz_max_intensity=1
	
		
	return(mz_points,mz_data,fit_mz_data,envelope_mz_data,mz_e,mz_max,mz_max_range,mz_max_intensity,mz_peaks,mz_noise,success,msg)

# ---------------------------------------------------------------------------
# Ocena sensownosci pozycji
def evaluate_position(spectrum,rt_max,isodistrib,charge,full_mz_margin,mz_correction,mz_margin,scans_margin):
	mz_e=1
	msg=''
	mz_e_success=1
	
	#polozenie piku monoizotopowego
	isopeaks_mz=isodistrib.get_peaks_mz(charge)+mz_correction
	mono_isopeak_mz=isopeaks_mz[0]
	
	#pobranie danych
	mz_range=isodistrib.get_envelope_mz_abs_range(charge,full_mz_margin,mz_correction)
	(mz_data,mz_points)=get_mz_data(spectrum,mz_range,rt_max,scans_margin)

	# wyznaczanie pikow
	(mz_peaks,mz_noise)=mspeak.find_peaks(mz_data,mz_points,filter_type=msfilter.FILTER_SG,filter_span=5,filter_order=2,noise_estimate_type=mspeak.NOISE_MAD,trim_peaks=True,trim_threshold=-1,np=True)
	if mz_peaks!=None and mz_peaks.shape[1]>0:
		(mpeak,mpeak_mz,mpeak_index)=get_data_range(mz_peaks[1,:],mz_peaks[0,:],[mono_isopeak_mz-mz_margin[0],mono_isopeak_mz+mz_margin[1]])

		if mpeak_index!=None:
			if len(mpeak)>1:
				(mpeak,mpeak_mz,ii)=get_data_max(mpeak,mpeak_mz)

			(ppeak,ppeak_mz,ppeak_index)=get_data_range(mz_peaks[1,:],mz_peaks[0,:],[mpeak_mz-1/float(charge)-mz_margin[0],mpeak_mz-1/float(charge)+mz_margin[1]])

			if ppeak_index!=None:
				if len(ppeak)>1:
					(ppeak,ppeak_mz,ii)=get_data_max(ppeak,ppeak_mz)
				if ppeak>0.8*mpeak:
					mz_e=-1
					msg='\n WARNING: additional peak detected' # in m/z range %f - %f' % (mz_range[0],mz_range[1])
	else:
		msg='\n ERROR: no peaks detected in m/z range %f - %f' % (mz_range[0],mz_range[1])
		mz_e_success=0
	
	return(mz_e,mz_data,mz_points,mz_peaks,mz_noise,mz_e_success,msg)

# ---------------------------------------------------------------------------
# Liczenie fitu
def mz_fit(data,mz_points,isopeaks,rp,mz_correction,mz_margin,max_intensity):
	msg=''

	try:
		init_p,fit_order=mz_fit_init_params(isopeaks,rp,mz_correction,max_intensity)
		fit_p,cov_x,infodict,mesg,success=scipy.optimize.leastsq(mz_fit_residuals,init_p,args=(mz_points,data),full_output=1,warning=False)
	except (ValueError):
		msg=msg+'\n ERROR: m/z fitting failed: ValueError in function leastsq' 
		return([],[],[],[],[],-1,-1,-1,msg)
	
	if success>4:
		msg=msg+'\n ERROR: m/z fitting failed: %s' % mesg 
		return([],[],[],[],[],-1,-1,-1,msg)
	
	if numpy.any(numpy.isnan(fit_p))>0 or numpy.any(numpy.isinf(fit_p))>0:
		msg=msg+'\n ERROR: m/z fit has INF or NAN values'
		return([],[],[],[],[],-1,-1,-2,msg)

	fit_probs,fit_mzs,fit_sigmas=mz_fit_peaks_form_params(fit_p)

	#ta ocena powinna byc gdzie indziej?
	fit_max_index=numpy.argmin(fit_order)
	mz_delta_ppm=mssputil.da2ppm(fit_mzs[fit_max_index]-(isopeaks[1,0]+mz_correction),isopeaks[1,0])
	if mz_delta_ppm<-1*mz_margin[0] or mz_delta_ppm>mz_margin[1]:
		msg=msg+('\n ERROR: m/z fit delta %0.3f ppm is not in range %0.3f - %0.3f' % (mz_delta_ppm,-1*mz_margin[0],mz_margin[1]))
		return([],[],[],[],[],-1,-1,-3,msg)

	fit_peaks_count=len(numpy.nonzero(fit_sigmas>0.0)[0])
	
	#poprawienie ujemnych
	if fit_peaks_count!=len(fit_sigmas):
		msg=msg+('\n WARNING: m/z fit has negative values')
		fit_sigmas=numpy.abs(fit_sigmas)
			
	fit_p=mz_fit_params_from_peaks(fit_probs,fit_mzs,fit_sigmas)

	return(fit_p,fit_order,fit_probs,fit_mzs,fit_sigmas,fit_max_index,fit_peaks_count,success,msg)

# ---------------------------------------------------------------------------
# 
def force_mz_fit(data,mz_points,isopeaks,rp,mz_correction,rp_mz_margin,max_intensity):
	msg=''

	fit_p,fit_order=mz_fit_init_params(isopeaks,rp,mz_correction,max_intensity)
	success=4
	
	#to powinno byc oparte o detekcje pikow
	fit_probs,fit_mzs,fit_sigmas=mz_fit_peaks_form_params(fit_p)
	for i in range(1,len(fit_probs)):
		(mz_max_intensity,mz_max,mz_max_index)=get_data_max(data,mz_points,(fit_mzs[i]-rp_mz_margin/2,fit_mzs[i]+rp_mz_margin/2))
		fit_probs[i]=mz_max_intensity
		
	#ta ocena powinna byc gdzie indziej?
	fit_max_index=numpy.argmin(fit_order)
	#mz_delta_ppm=mssputil.da2ppm(fit_mzs[fit_max_index]-(isopeaks[1,0]+mz_correction),isopeaks[1,0])
	#if mz_delta_ppm<-1*mz_margin[0] or mz_delta_ppm>mz_margin[1]:
	#	msg=msg+('\n ERROR: m/z fit delta %0.3f ppm is not in range %0.3f - %0.3f' % (mz_delta_ppm,-1*mz_margin[0],mz_margin[1]))
	#	return([],[],[],[],[],-1,-1,-3,msg)

	fit_peaks_count=len(numpy.nonzero(fit_sigmas>0.0)[0])
	print fit_peaks_count
	fit_p=mz_fit_params_from_peaks(fit_probs,fit_mzs,fit_sigmas)
	
	
	return(fit_p,fit_order,fit_probs,fit_mzs,fit_sigmas,fit_max_index,fit_peaks_count,success,msg)

# ---------------------------------------------------------------------------
# Zamiana parametry -> (wysokosci, pozycja, szerokosci)
def mz_fit_peaks_form_params(fit_params):

	peaks_count=len(fit_params)/MS_MZ_FIT_PARAMS

	probs=numpy.zeros(peaks_count)
	masses=numpy.zeros(peaks_count)
	sigmas=numpy.zeros(peaks_count)

	for i in range(0,peaks_count):
		min_index=MS_MZ_FIT_PARAMS*i

		probs[i]=fit_params[min_index]
		masses[i]=fit_params[min_index+1]
		sigmas[i]=fit_params[min_index+2]

	return probs,masses,sigmas

# ---------------------------------------------------------------------------
# Zamiana (wysokosci, pozycja, szerokosci) -> parametry
def mz_fit_params_from_peaks(probs,mzs,sigmas):

	peaks_count=min(min(len(probs),len(mzs)),len(sigmas))

	p=numpy.zeros(MS_MZ_FIT_PARAMS*peaks_count)
	for i in range(0,peaks_count):
		min_index=MS_MZ_FIT_PARAMS*i
		p[min_index]=probs[i]
		p[min_index+1]=mzs[i]
		p[min_index+2]=sigmas[i]

	return p

# ---------------------------------------------------------------------------
#  Obliczanie poczatkowych parametorw fitu
def mz_fit_init_params(isopeaks,rp,mz_correction=0,scale=1):

	index=numpy.argsort(isopeaks[0,:])
	mzs=isopeaks[1,index]+mz_correction
	probs=isopeaks[2,index]*scale
	sigmas=mzs/(rp*2*numpy.sqrt(2*numpy.log(2.0)))

	p=mz_fit_params_from_peaks(probs,mzs,sigmas)
	
	return p,index

# ---------------------------------------------------------------------------
# Funkcja fitu
def mz_fit_function(p,mz_points):

	peaks_count=p.shape[0]/3
	values=numpy.zeros(mz_points.shape[0])

	for i in range(0,peaks_count):
		min_index=3*i
		values=values+p[min_index]*numpy.exp(-0.5*((mz_points-p[min_index+1])**2)/(p[min_index+2]**2))
		
	return values

# ---------------------------------------------------------------------------
# Reszy modelu
def mz_fit_residuals(p,mz_points,data):
	
	err=mz_fit_function(p,mz_points)-data

	return(err)

# ---------------------------------------------------------------------------
#
def mz_fit_area(p):

	peaks_count=len(p)/3
	c=numpy.sqrt(2*numpy.pi)

	mz_area=0
	for i in range(0,peaks_count):
		mz_area=mz_area+p[i*3]*p[i*3+2]*c
	return mz_area

#
# Fitowanie po RT
#

# --------------------------------------------------------------------------
# Tworzenie i  ocena fitu RT
def create_rt_fit(spectrum,fit,rt_points,rt_traces,mz_points,rt_max,rt_max_index):
	msg=''
	success=1

	#wybranie sladu odpowiadajacemu max m/z
	#rt_data=rt_traces.max(1)
	delta=numpy.abs(mz_points-fit.mz_pos)
	rt_data=rt_traces[:,numpy.argmin(delta)]

	#ochrona przed nietrafieniem w pozycje max m/z 
	if rt_data.max()==0:
		rt_data=rt_traces.max(1)
		msg=msg+'\n WARNING: using max in range'

	#detekcja pikow
#	(rt_peaks,rt_noise)=mspeak.find_peaks(rt_data,rt_points,filter_type=msfilter.FILTER_SG,filter_span=5,filter_order=2,noise_estimate_type=mspeak.NOISE_MAD,trim_peaks=False,trim_threshold=0,np=True)
	(rt_peaks,rt_noise)=mspeak.find_peaks2(rt_data,rt_points)

	#ograniczenie zakresu fitu do obszaru piku najblizszego rt_max
	if rt_peaks!=None and rt_peaks.shape[1]>0:
		delta=numpy.fabs(rt_peaks[0,:]-rt_max)
		peak_index=numpy.argmin(delta)
	
		rt_fit_range=(rt_peaks[5,peak_index],min(rt_peaks[8,peak_index]+1,len(rt_points)))
		if rt_fit_range[1]-rt_fit_range[0]>=3:
			peak_rt_max=rt_peaks[0,peak_index]
			peak_rt_max_intensity=rt_peaks[1,peak_index]
		else:
			rt_fit_range=[]
			msg=msg+'\n WARNING: less than 3 points in RT peak range (left %f, rigth %f, max %f)' % (rt_peaks[5,peak_index],rt_peaks[8,peak_index],rt_peaks[1,peak_index])
	else:
		rt_fit_range=[]
		msg=msg+'\n WARNING: no RT peaks detected'

	if len(rt_fit_range)==0:
		rt_fit_range=(0,len(rt_points))
		peak_rt_max=rt_max
		peak_rt_max_intensity=rt_data[rt_max_index]


	msg=msg+'\n RT joint chromatogram peak range: %f - %f (%d points)' % (-rt_points[0],-rt_points[-1],len(rt_points))
	msg=msg+'\n RT joint chromatogram peak max position: %f' % (-rt_max)
	msg=msg+'\n RT fit range: %f - %f (%d points)' % (-rt_points[rt_fit_range[0]],-rt_points[min(rt_fit_range[1],len(rt_points)-1)],len(rt_points[rt_fit_range[0]:rt_fit_range[1]]))
	msg=msg+'\n RT fit range max position: %f' % (-peak_rt_max)
	msg=msg+'\n RT diff: %f' % (rt_max-peak_rt_max)

	rt_fit_p,rt_succes,rt_msg=rt_fit(rt_data[rt_fit_range[0]:rt_fit_range[1]],rt_points[rt_fit_range[0]:rt_fit_range[1]],peak_rt_max,peak_rt_max_intensity,native=True)
	msg=msg+rt_msg

	fit.rt_pos=numpy.around(rt_max,6)
	fit.rt_params=numpy.around(rt_fit_p,6)
	fit.rt_flag=rt_succes

	rt_fit_data=evaluate_rt_fit(fit,rt_data[rt_fit_range[0]:rt_fit_range[1]],rt_points[rt_fit_range[0]:rt_fit_range[1]])

	msg=msg+('\n RT fit max: %f\n RT fit delta: %f\n RT fit R-square: %f' % (rt_fit_p[1],(-rt_max-rt_fit_p[1]),fit.rt_rsquare))
	
	#przy obecnych zabezpieczeniach (zamianie na dodatnie) to ostrzezenie nie powinno sie pojawiac
	if fit.rt_area<0:
		msg=msg+'\n WARNING: RT area negative'

	return(rt_points,rt_data,rt_fit_data,rt_fit_range,rt_max,rt_peaks,rt_noise,success,msg)

# ---------------------------------------------------------------------------
# Obliczanie fitu RT
def rt_fit(data,rt_points,rt_max,rt_max_intensity,rt_fit_type=RT_FIT_GAUSS,native=True):

	msg=''

	if native==True and rt_points[0]<0:
		rt_points=-1*rt_points
		rt_max=-1*rt_max
		
	try:
		init_p=rt_fit_params(rt_max,rt_max_intensity,rt_fit_type)
		fit_p,cov_x,infodict,mesg,success=scipy.optimize.leastsq(rt_fit_residuals,init_p,args=(rt_points,data,rt_fit_type),full_output=1,warning=False)
	except (ValueError):
		#zamienic na error
		msg=msg+'\n ERROR: RT fitting failed: ValueError in function leastsq' 
	
	#zamienic na error
	if success>4:
		msg=msg+'\n WARNING: RT fitting failed: %s' % mesg 

	#poprawienie ujemnych		
	if len(numpy.nonzero(fit_p<0.0)[0])>0:
		msg=msg+('\n WARNING: RT fit has negative sigmas')
		fit_p=numpy.abs(fit_p)
		
	if rt_fit_type==RT_FIT_GAUSS:
		fit_p=numpy.hstack([fit_p,RT_FIT_GAUSS])
	
	return(fit_p,success,msg)

# ---------------------------------------------------------------------------
# Obliczanie poczatkowych prarametrow fitu RT
def rt_fit_params(rt_max,rt_max_intensity,rt_fit_type=RT_FIT_GAUSS):	#dodac szacowana szerokosc

	if rt_fit_type==RT_FIT_EMG:
		# p[0] - wysokosc A = (2*w*sqrt(pi/2))*h
		# p[1] - centroid z
		# p[2] - szerokosc w
		# p[3] - skrosnosc s
		p=numpy.zeros(4)
		p[0]=rt_max_intensity
		p[1]=rt_max
		p[2]=2
		p[3]=1

	elif rt_fit_type==RT_FIT_GAUSS:
		# p[0] - wysokosc
		# p[1] - centroid
		# p[2] - szerokosc
		p=numpy.zeros(3)
		p[0]=rt_max_intensity
		p[1]=rt_max
		p[2]=0.4		#to powinno byc uzaleznione od parametrow widma
	return(p)

# ---------------------------------------------------------------------------
# Funkcja fitu RT
def rt_fit_function(p,rt_points,rt_fit_type=RT_FIT_GAUSS):

	if rt_fit_type==RT_FIT_GAUSS:
		values=p[0]*numpy.exp(-0.5*((rt_points-p[1])**2)/(p[2]**2))
	else:
		values=(p[0]/(2*p[3])) * numpy.exp(0.5*(p[2]/p[3])**2-(rt_points-p[1])/p[3]) * (1-(scipy.special.erf((p[2]/p[3]-(rt_points-p[1])/p[2])/numpy.sqrt(2))))

	return values

# ---------------------------------------------------------------------------
# Reszy fitu RT
def rt_fit_residuals(p,rt_points,data,rt_fit_type=RT_FIT_GAUSS):

	err=rt_fit_function(p,rt_points,rt_fit_type)-data

	return(err)


# --------------------------------------------------------------------------
#
def get_rt_trace(spectrum,rt_range,mz,mz_correction=0,mz_margin=[20,20],rt=None):

	right_min=rt_range[0]
	left_min=rt_range[1]
	peak_max=rt_range[2]

	if rt==None:
		rt_step=mssputil.get_rt_step(spectrum)
		add_points=4-(right_min-left_min)/rt_step
		if add_points>0:
			left_min=left_min-numpy.ceil(add_points/2)*rt_step
			right_min=right_min+numpy.ceil(add_points/2)*rt_step
		rt=mssputil.get_rt_in_range(spectrum,[right_min,left_min],native=True)

	#zakres taki sam jak przy tworzeniu chromatogramu
	mz_range=[mz+mz_correction-mz_margin[0]*mz/(1e+6),mz+mz_correction+mz_margin[1]*mz/(1e+6)]

	if spectrum.domain==mssputil.DOMAIN_FREQ:
		mz_points=mssputil.get_f_in_range(spectrum,mssputil.mz2freq_list(mz_range,spectrum))
	else:
		mz_points=mssputil.get_mz_in_range(spectrum,mz_range,native=True)

	#dopasowanie nastapi do maksymalnych wartosci w zakresie m/z => czy nie lepiej mean???
	rt_traces=mssputil.get_datalines(spectrum,mz_points,rt)
	rt_trace=rt_traces.max(1)

	#ograniczenie szukania maximum w zakresie +/-3skanow od maximum piku chromatogramu
	#!!!!!powinno byc sprawdzane czy to faktycznie max lub detakcja pikow i wybor najblizsego
	peak_max_range=(rt>=peak_max-3*rt_step) * (rt<=peak_max+3*rt_step)
	rt_max_index=numpy.argmax(rt_trace[peak_max_range])+numpy.nonzero(peak_max_range)[0][0]
	rt_max=rt[rt_max_index]

	return rt,rt_trace,rt_max,rt_max_index,rt_traces,mz_points

#
# Bledy fitu
#

# ---------------------------------------------------------------------------
#
def evaluate_mz_fit(fit,isodistrib,charge,rp,mz_data,mz_points,fit_probs,fit_mzs,fit_sigmas):

	# m/z
	fit_mz_data=mz_fit_function(fit.mz_params,mz_points)
	fit_mz_data=numpy.maximum(fit_mz_data,numpy.zeros(len(mz_points)))

	fit.mz_area=mz_fit_area(fit.mz_params)

	fit.mz_rsquare,fit.mz_rmse=msmath.rsquare(mz_data,fit_mz_data,compute_rmse=True)
	fit.mz_rmse=fit.mz_rmse*numpy.sqrt(len(mz_data))/fit.mz_area

	# Envelope
	envelope_mz_data,envelope_mz_points=isodistrib.get_envelope_for_points(mz_points,charge,rp,fit.mz_delta,fit_probs.max())

	fit.mz_env_rsquare,fit.mz_env_rmse=msmath.rsquare(fit_mz_data,envelope_mz_data,compute_rmse=True)
	
	fit.mz_env_rsquare=max(fit.mz_env_rsquare,0)
	fit.mz_env_rmse=fit.mz_env_rmse*numpy.sqrt(len(mz_data))/fit.mz_area

	return fit_mz_data,envelope_mz_data

# ---------------------------------------------------------------------------
#
def evaluate_rt_fit(peak_fit,rt_data=None,rt_points=None,rt_fit_type=RT_FIT_GAUSS):

	# RT
	fit_rt_data=rt_fit_function(peak_fit.rt_params,-1*rt_points,rt_fit_type=rt_fit_type)
	peak_fit.rt_area=scipy.integrate.trapz(fit_rt_data,-1*rt_points)

	peak_fit.rt_rsquare,peak_fit.rt_rmse=msmath.rsquare(rt_data,fit_rt_data,compute_rmse=True)
	peak_fit.rt_rmse=peak_fit.rt_rmse*numpy.sqrt(len(rt_data))/peak_fit.rt_area

	return fit_rt_data

# ---------------------------------------------------------------------------
#
def get_data_range(data,points,points_range):

	range_index=None
	range_data=None
	range_points=None
	if data!=None and points!=None and points_range!=None:
		range_index=((points>points_range[0]) & (points<points_range[1]))
		if numpy.any(range_index)==True:
			range_data=data[range_index]
			range_points=points[range_index]
		else:
			range_index=None

	return(range_data,range_points,range_index)

# ---------------------------------------------------------------------------
#
def get_data_max(data,points,points_range=None):

	max_data=numpy.nan
	max_index=numpy.nan
	max_point=numpy.nan
	
	if points_range==None:
		max_index=numpy.argmax(data)
		max_data=data[max_index]
		max_point=points[max_index]
	else:
		(range_data,range_points,range_index)=get_data_range(data,points,points_range)
		if range_index!=None:
			max_index=numpy.argmax(range_data)
			max_data=range_data[max_index]
			max_point=range_points[max_index]

	return(max_data,max_point,max_index)

# ---------------------------------------------------------------------------
#
def get_mz_data(spectrum,mz_range,rt,scans_margin=0):

	if spectrum.domain==mssputil.DOMAIN_FREQ:
		f_range=mssputil.mz2freq_list(mz_range,spectrum)
		f_points=mssputil.get_f_in_range(spectrum,f_range)

		mz_points=mssputil.freq2mz_list(f_points,spectrum)
		mz_data=mssputil.get_dataline(spectrum,f_points,-1*rt)
	else:
		mz_points=numpy.array(mssputil.get_mz_in_range(spectrum,mz_range,native=False))
		mz_data=numpy.array(mssputil.get_dataline(spectrum,-1*mz_points,-1*rt))

		rt_step=mssputil.get_rt_step(spectrum)

		if scans_margin!=0:
			for i in range(1,scans_margin+1):
				mz_data=mz_data+numpy.array(mssputil.get_dataline(spectrum,-1*mz_points,-1*(rt-i*rt_step)))
			for i in range(1,scans_margin+1):
				mz_data=mz_data+numpy.array(mssputil.get_dataline(spectrum,mz_points,-1*(rt+i*rt_step)))
			mz_data=mz_data/(scans_margin+1)
	return(mz_data,mz_points)

#
# Rysowanie fitu
# ---------------------------------------------------------------------------
#
def plot_chromatogram_fit(session,chromat,peak_nr,mz_points,mz_data,mz_fit_data,mz_envelope_data,rt_points,rt_data,rt_fit_data,rt_fit_range,title):

	yformatter=matplotlib.ticker.ScalarFormatter()
	yformatter.set_scientific(True)
	yformatter.set_powerlimits((0,0))

	fig=Figure(figsize=(8,10))
	p=fig.add_subplot(311)
	p.plot(-1*chromat.rt,chromat.data,color='b')

	peak=chromat.peaks[peak_nr]
	p.plot(-1*chromat.rt[peak[1]:(peak[2]+1)],chromat.data[peak[1]:(peak[2]+1)],color='r')


	p.set_xlim([-1*chromat.rt[0],-1*chromat.rt[-1]])
	p.set_title("Joint chromatogram")
	p.set_xlabel("RT [min.]")

	p.yaxis.set_major_formatter(yformatter)
	p.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1f'))
	for label in p.yaxis.get_ticklabels() + p.xaxis.get_ticklabels():
		label.set_fontsize(10)
	p.set_position([0.07,0.71,0.9,0.25])

	if rt_points!=None:
		p=fig.add_subplot(312)
		p.plot(-1*rt_points,rt_data,color='b')
		if rt_fit_data!=None and len(rt_fit_data)>0:
			p.plot(-1*rt_points[rt_fit_range[0]:rt_fit_range[1]],rt_fit_data,color='g')

		p.set_xlim([-1*rt_points[0],-1*rt_points[-1]])
		p.set_title("Monoisotopic peak chromatogram in selected RT range")
		p.set_xlabel("RT [min.]")
		p.yaxis.set_major_formatter(yformatter)
		p.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1f'))
		for label in p.yaxis.get_ticklabels() + p.xaxis.get_ticklabels():
			label.set_fontsize(10)
		p.set_position([0.07,0.38,0.9,0.25])

	if mz_points!=None:
		p=fig.add_subplot(313)
		p.plot(mz_points,mz_data,color='b')
		if mz_fit_data!=None and len(mz_fit_data)>0:
			p.plot(mz_points,mz_fit_data,color='g')
		if mz_envelope_data!=None and len(mz_envelope_data)>0:
			p.plot(mz_points,mz_envelope_data,color='r')

		p.set_xlim([mz_points[0],mz_points[-1]])
		p.set_title("m/z profile")
		p.set_xlabel("m/z")
		p.yaxis.set_major_formatter(yformatter)
		p.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.4f'))
		for label in p.yaxis.get_ticklabels() + p.xaxis.get_ticklabels():
			label.set_fontsize(10)
		p.set_position([0.07,0.05,0.9,0.25])

	d=mstkutil.PlotDialog(session,fig,title)
	d.show_window(1)
