import numpy
import scipy
import scipy.stats.mstats

import msmath
import math
import mstkutil
import mssputil

NORM_STR=('None','Total counts','User defined','Global median','Global mean','Global gmean','Protein median','Protein mean','Protein gmean','RLS','LOWESS')
NORM_NONE=0
NORM_TOTAL_COUNTS=1
NORM_USER=2
NORM_GLOBAL_MEDIAN=3
NORM_GLOBAL_MEAN=4
NORM_GLOBAL_GMEAN=5
NORM_PROTEIN_MEDIAN=6
NORM_PROTEIN_MEAN=7
NORM_PROTEIN_GMEAN=8
NORM_RLS=9
NORM_LOWESS=10

# --------------------------------------------------------------------------
#
def get_normalization_type(name):
	filter_type=-1
	for i in range(0,len(NORM_STR)):
		if name.lower()==NORM_STR[i].lower():
			filter_type=i
			break
	return filter_type

# ----------------------------------------------------------------------------
#
class normalization_menu(mstkutil.option_menu):

	def __init__(self,parent,label,allow_no_choice=0):
		self.allow_no_choice=allow_no_choice

		mstkutil.option_menu.__init__(self,parent,label,NORM_STR,NORM_STR[0])

	# --------------------------------------------------------------------------
	#
	def get_value(self):
		return get_normalization_type(self.get())
	
	# --------------------------------------------------------------------------
	#
	def set_value(self,value):
		entry='None'
		if value>=0 and value <len(NORM_STR):
			entry=NORM_STR[value]
		entry=self.set(entry)

# ----------------------------------------------------------------------------
#
def global_normalization(x,y,norm_method=NORM_GLOBAL_MEDIAN,norm_factor=1.0):

	index_x=numpy.isfinite(x)
	index_y=numpy.isfinite(y)

	if len(index_x)>0 and len(index_y)>0:
		if norm_method==NORM_GLOBAL_MEDIAN:
			mx=numpy.median(x[index_x])
			my=numpy.median(y[index_y])
			norm_factor=mx/my
		elif norm_method==NORM_GLOBAL_MEAN:
			mx=numpy.mean(x[index_x])
			my=numpy.mean(y[index_y])
			norm_factor=mx/my
		elif norm_method==NORM_GLOBAL_GMEAN:
			mx=scipy.stats.mstats.gmean(x[index_x])
			my=scipy.stats.mstats.gmean(y[index_y])
			norm_factor=mx/my

		for i in range(0,len(y)):
			y[i]=y[i]*norm_factor

# ----------------------------------------------------------------------------
#

def get_protein_norm_factor(x,y,norm_method,peaks_list,protein):
	norm_factor=1.0

	protein=protein.strip()
	if protein!=None and len(protein)>0:
		protein=protein.lower()
		protein_x=[]
		protein_y=[]
		for i in range(0,len(peaks_list)):
			if mssputil.get_protein(peaks_list[i]).lower()==protein:
				protein_x.append(x[i])
				protein_y.append(y[i])

		if len(protein_x)>0:
			protein_x=numpy.array(protein_x)
			protein_y=numpy.array(protein_y)

			index_x=numpy.isfinite(protein_x)
			index_y=numpy.isfinite(protein_y)

			if len(index_x)>0 and len(index_y)>0:
				if norm_method==NORM_PROTEIN_MEDIAN:
					mx=numpy.median(protein_x[index_x])
					my=numpy.median(protein_y[index_y])
					norm_factor=mx/my
				elif norm_method==NORM_PROTEIN_MEAN:
					mx=numpy.mean(protein_x[index_x])
					my=numpy.mean(protein_y[index_y])
					norm_factor=mx/my
				elif norm_method==NORM_PROTEIN_GMEAN:
					mx=scipy.stats.mstats.gmean(protein_x[index_x])
					my=scipy.stats.mstats.gmean(protein_y[index_y])
					norm_factor=mx/my
	return(norm_factor)

# ----------------------------------------------------------------------------
#
def lowess_normalization(x,y,session,logscale,f,nsteps,delta):

	if logscale==True:
		points_x=numpy.log10(x)
		points_y=numpy.log10(y)
	else:
		points_x=x
		points_y=y
	
	idx=numpy.isfinite(points_x)*numpy.isfinite(points_y)
	if numpy.sum(idx)>0:
		points_x=points_x[idx]
		points_y=points_y[idx]
		
		
		sort_idx=numpy.argsort(points_x)
		points_y=points_y[sort_idx]
		points_x=points_x[sort_idx]
		e=points_y-points_x
		
		fit,inliers_index=msmath.lowess(session,points_x,e,f,nsteps,delta,0)

		a=numpy.zeros(len(points_x))
		b=numpy.zeros(len(points_x))
		for i in range(0,len(points_x)):
			a[sort_idx[i]]=points_y[i]
			b[sort_idx[i]]=fit[i]
		
#		y[idx]=a-b
#		if logscale==True:
#			for i in range(0,len(y)):
#				if numpy.isfinite(y[i])==1 and numpy.isfinite(x[i])==1:
#					y[i]=math.pow(10,y[i])
		counter=0
		for i in range(0,len(y)):
			if numpy.isfinite(y[i])==1 and numpy.isfinite(x[i])==1:
				y[i]=a[counter]-b[counter]
				if logscale==True:
					y[i]=math.pow(10,y[i])
				counter=counter+1
#		#	elif numpy.isfinite(y[i])==1:
#		#		y[i]=numpy.nan

	#else:
	#	for i in range(0,len(y)):
	#		if numpy.isfinite(y[i])==1:
	#			y[i]=y[i]-msmath.lowess_val(points_y,fit,y[i])
			
# ----------------------------------------------------------------------------
#
def get_lowess_fit(x,y,session,f,nsteps,delta):
	
	idx=numpy.isfinite(x)*numpy.isfinite(y)
	if numpy.sum(idx)>0:
		points_x=points_x[idx]
		points_y=points_y[idx]

		sort_idx=numpy.argsort(points_x)
		points_x=points_x[sort_idx]
		points_y=points_y[sort_idx]

		fit,inliers_index=msmath.lowess(session,points_x,points_y,f,nsteps,delta,0)
		fit_x=points_x
